#include "lcd.h"
#include <avr/io.h> // for AVR hardware
#include <avr/cpufunc.h> // for _NOP
#include <util/delay.h> // for delays

// initialize pin states
inline void lcd_init_pins()
{
    LCD_DDR |= _BV(0) |
               _BV(1) |
               _BV(2) |
               _BV(3) |
               _BV(LCD_PIN_E) |
               _BV(LCD_PIN_RS) |
               _BV(LCD_PIN_LED);
    LCD_PORT = 0;
}

// send a byte to the LCD, RS pin must be setup before calling this function
static inline void lcd_sendbyte(uint8_t data)
{
    // setup the data bits (first 4 bits)
    LCD_PORT = (LCD_PORT & 0xF0) | (0x0F & ((data & 0xF0) >> 4));
    _NOP(); _NOP(); _NOP(); // some delay
    
    // raise the E pin
    LCD_PORT |= _BV(LCD_PIN_E);
    _NOP(); _NOP(); _NOP(); // some delay
    
    // lower the E pin, therefor inputting the data
    LCD_PORT &= ~_BV(LCD_PIN_E);
    _NOP(); _NOP(); _NOP(); // some delay
    
    // setup the data bits (last 4 bits)
    LCD_PORT = (LCD_PORT & 0xF0) | (0x0F & data);
    _NOP(); _NOP(); _NOP(); // some delay
    
    // raise the E pin
    LCD_PORT |= _BV(LCD_PIN_E);
    _NOP(); _NOP(); _NOP(); // some delay
    
    // lower the E pin, therefor inputting the data
    LCD_PORT &= ~_BV(LCD_PIN_E);
    
    // minimum time for execution
    _delay_us(50);
}

// display a character on the LCD
void lcd_sendchar(char c)
{
    // displaying a character means we turn on RS
    LCD_PORT |= _BV(LCD_PIN_RS);
    
    lcd_sendbyte(c);
}

// send a command to the LCD
void lcd_sendcmd(uint8_t data)
{
    // command means we turn off RS
    LCD_PORT &= ~_BV(LCD_PIN_RS);
    
    lcd_sendbyte(data);
}

// sets the position of the cursor, x is 0 indexed, line is not
void lcd_setpos(uint8_t x, uint8_t line)
{
    if (line == 1)
    {
        lcd_sendcmd(0b10000000 | 0x00 | x);
    }
    else if (line == 2)
    {
        lcd_sendcmd(0b10000000 | 0x40 | x);
    }
    
    // TODO support for 4 line displays needs to be added
}

// initialize the LCD
void lcd_init()
{
    // setup the pins
    lcd_init_pins();
    
    // clear the display
    lcd_sendcmd(0b00000001);
    _delay_us(1700); // requires a delay
    // return home
    lcd_sendcmd(0b00000010);
    _delay_us(1700); // requires a delay
    // these two effectively resets the display
    
    // function set, enable 4 bit mode first
    lcd_sendcmd(0b00101000);
    
    // entry mode set: left to right, no display shift
    //lcd_sendcmd(0b00000110); // commented out because it's already default
    
    // display control: on, no cursor, no cursor blink
    lcd_sendcmd(0b00001100);
    
    // start at line 1 position 0
    lcd_setpos(0, 1);
}